# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.async_support.base.exchange import Exchange
from ccxt.abstract.oceanex import ImplicitAPI
from ccxt.base.types import Balances, Int, Market, Num, Order, OrderBook, OrderSide, OrderType, Str, Strings, Ticker, Tickers, Trade, TradingFees
from typing import List
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import AuthenticationError
from ccxt.base.errors import PermissionDenied
from ccxt.base.errors import ArgumentsRequired
from ccxt.base.errors import BadRequest
from ccxt.base.errors import InsufficientFunds
from ccxt.base.errors import InvalidOrder
from ccxt.base.errors import OrderNotFound
from ccxt.base.decimal_to_precision import TICK_SIZE


class oceanex(Exchange, ImplicitAPI):

    def describe(self):
        return self.deep_extend(super(oceanex, self).describe(), {
            'id': 'oceanex',
            'name': 'OceanEx',
            'countries': ['BS'],  # Bahamas
            'version': 'v1',
            'rateLimit': 3000,
            'urls': {
                'logo': 'https://user-images.githubusercontent.com/1294454/58385970-794e2d80-8001-11e9-889c-0567cd79b78e.jpg',
                'api': {
                    'rest': 'https://api.oceanex.pro',
                },
                'www': 'https://www.oceanex.pro.com',
                'doc': 'https://api.oceanex.pro/doc/v1',
                'referral': 'https://oceanex.pro/signup?referral=VE24QX',
            },
            'has': {
                'CORS': None,
                'spot': True,
                'margin': False,
                'swap': None,  # has but unimplemented
                'future': None,
                'option': None,
                'cancelAllOrders': True,
                'cancelOrder': True,
                'cancelOrders': True,
                'createMarketOrder': True,
                'createOrder': True,
                'fetchBalance': True,
                'fetchBorrowRateHistories': False,
                'fetchBorrowRateHistory': False,
                'fetchClosedOrders': True,
                'fetchCrossBorrowRate': False,
                'fetchCrossBorrowRates': False,
                'fetchDepositAddress': False,
                'fetchDepositAddresses': False,
                'fetchDepositAddressesByNetwork': False,
                'fetchIsolatedBorrowRate': False,
                'fetchIsolatedBorrowRates': False,
                'fetchMarkets': True,
                'fetchOHLCV': True,
                'fetchOpenOrders': True,
                'fetchOrder': True,
                'fetchOrderBook': True,
                'fetchOrderBooks': True,
                'fetchOrders': True,
                'fetchTicker': True,
                'fetchTickers': True,
                'fetchTime': True,
                'fetchTrades': True,
                'fetchTradingFee': False,
                'fetchTradingFees': True,
                'fetchTransactionFees': None,
            },
            'timeframes': {
                '1m': '1',
                '5m': '5',
                '15m': '15',
                '30m': '30',
                '1h': '60',
                '2h': '120',
                '4h': '240',
                '6h': '360',
                '12h': '720',
                '1d': '1440',
                '3d': '4320',
                '1w': '10080',
            },
            'api': {
                'public': {
                    'get': [
                        'markets',
                        'tickers/{pair}',
                        'tickers_multi',
                        'order_book',
                        'order_book/multi',
                        'fees/trading',
                        'trades',
                        'timestamp',
                    ],
                    'post': [
                        'k',
                    ],
                },
                'private': {
                    'get': [
                        'key',
                        'members/me',
                        'orders',
                        'orders/filter',
                    ],
                    'post': [
                        'orders',
                        'orders/multi',
                        'order/delete',
                        'order/delete/multi',
                        'orders/clear',
                    ],
                },
            },
            'fees': {
                'trading': {
                    'tierBased': False,
                    'percentage': True,
                    'maker': self.parse_number('0.001'),
                    'taker': self.parse_number('0.001'),
                },
            },
            'commonCurrencies': {
                'PLA': 'Plair',
            },
            'precisionMode': TICK_SIZE,
            'exceptions': {
                'codes': {
                    '-1': BadRequest,
                    '-2': BadRequest,
                    '1001': BadRequest,
                    '1004': ArgumentsRequired,
                    '1006': AuthenticationError,
                    '1008': AuthenticationError,
                    '1010': AuthenticationError,
                    '1011': PermissionDenied,
                    '2001': AuthenticationError,
                    '2002': InvalidOrder,
                    '2004': OrderNotFound,
                    '9003': PermissionDenied,
                },
                'exact': {
                    'market does not have a valid value': BadRequest,
                    'side does not have a valid value': BadRequest,
                    'Account::AccountError: Cannot lock funds': InsufficientFunds,
                    'The account does not exist': AuthenticationError,
                },
            },
        })

    async def fetch_markets(self, params={}) -> List[Market]:
        """
        retrieves data on all markets for oceanex
        :see: https://api.oceanex.pro/doc/v1/#markets-post
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: an array of objects representing market data
        """
        request = {'show_details': True}
        response = await self.publicGetMarkets(self.extend(request, params))
        #
        #    {
        #        "id": "xtzusdt",
        #        "name": "XTZ/USDT",
        #        "ask_precision": "8",
        #        "bid_precision": "8",
        #        "enabled": True,
        #        "price_precision": "4",
        #        "amount_precision": "3",
        #        "usd_precision": "4",
        #        "minimum_trading_amount": "1.0"
        #    },
        #
        markets = self.safe_value(response, 'data', [])
        return self.parse_markets(markets)

    def parse_market(self, market) -> Market:
        id = self.safe_value(market, 'id')
        name = self.safe_value(market, 'name')
        baseId, quoteId = name.split('/')
        base = self.safe_currency_code(baseId)
        quote = self.safe_currency_code(quoteId)
        baseId = baseId.lower()
        quoteId = quoteId.lower()
        symbol = base + '/' + quote
        return {
            'id': id,
            'symbol': symbol,
            'base': base,
            'quote': quote,
            'settle': None,
            'baseId': baseId,
            'quoteId': quoteId,
            'settleId': None,
            'type': 'spot',
            'spot': True,
            'margin': False,
            'swap': False,
            'future': False,
            'option': False,
            'active': None,
            'contract': False,
            'linear': None,
            'inverse': None,
            'contractSize': None,
            'expiry': None,
            'expiryDatetime': None,
            'strike': None,
            'optionType': None,
            'precision': {
                'amount': self.parse_number(self.parse_precision(self.safe_string(market, 'amount_precision'))),
                'price': self.parse_number(self.parse_precision(self.safe_string(market, 'price_precision'))),
            },
            'limits': {
                'leverage': {
                    'min': None,
                    'max': None,
                },
                'amount': {
                    'min': None,
                    'max': None,
                },
                'price': {
                    'min': None,
                    'max': None,
                },
                'cost': {
                    'min': self.safe_number(market, 'minimum_trading_amount'),
                    'max': None,
                },
            },
            'created': None,
            'info': market,
        }

    async def fetch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        fetches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market
        :see: https://api.oceanex.pro/doc/v1/#ticker-post
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        market = self.market(symbol)
        request = {
            'pair': market['id'],
        }
        response = await self.publicGetTickersPair(self.extend(request, params))
        #
        #     {
        #         "code":0,
        #         "message":"Operation successful",
        #         "data": {
        #             "at":1559431729,
        #             "ticker": {
        #                 "buy":"0.0065",
        #                 "sell":"0.00677",
        #                 "low":"0.00677",
        #                 "high":"0.00677",
        #                 "last":"0.00677",
        #                 "vol":"2000.0"
        #             }
        #         }
        #     }
        #
        data = self.safe_dict(response, 'data', {})
        return self.parse_ticker(data, market)

    async def fetch_tickers(self, symbols: Strings = None, params={}) -> Tickers:
        """
        fetches price tickers for multiple markets, statistical information calculated over the past 24 hours for each market
        :see: https://api.oceanex.pro/doc/v1/#multiple-tickers-post
        :param str[]|None symbols: unified symbols of the markets to fetch the ticker for, all market tickers are returned if not assigned
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a dictionary of `ticker structures <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        symbols = self.market_symbols(symbols)
        if symbols is None:
            symbols = self.symbols
        marketIds = self.market_ids(symbols)
        request = {'markets': marketIds}
        response = await self.publicGetTickersMulti(self.extend(request, params))
        #
        #     {
        #         "code":0,
        #         "message":"Operation successful",
        #         "data": {
        #             "at":1559431729,
        #             "ticker": {
        #                 "buy":"0.0065",
        #                 "sell":"0.00677",
        #                 "low":"0.00677",
        #                 "high":"0.00677",
        #                 "last":"0.00677",
        #                 "vol":"2000.0"
        #             }
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', [])
        result = {}
        for i in range(0, len(data)):
            ticker = data[i]
            marketId = self.safe_string(ticker, 'market')
            market = self.safe_market(marketId)
            symbol = market['symbol']
            result[symbol] = self.parse_ticker(ticker, market)
        return self.filter_by_array_tickers(result, 'symbol', symbols)

    def parse_ticker(self, data, market: Market = None):
        #
        #         {
        #             "at":1559431729,
        #             "ticker": {
        #                 "buy":"0.0065",
        #                 "sell":"0.00677",
        #                 "low":"0.00677",
        #                 "high":"0.00677",
        #                 "last":"0.00677",
        #                 "vol":"2000.0"
        #             }
        #         }
        #
        ticker = self.safe_value(data, 'ticker', {})
        timestamp = self.safe_timestamp(data, 'at')
        symbol = self.safe_symbol(None, market)
        return self.safe_ticker({
            'symbol': symbol,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'high': self.safe_string(ticker, 'high'),
            'low': self.safe_string(ticker, 'low'),
            'bid': self.safe_string(ticker, 'buy'),
            'bidVolume': None,
            'ask': self.safe_string(ticker, 'sell'),
            'askVolume': None,
            'vwap': None,
            'open': None,
            'close': self.safe_string(ticker, 'last'),
            'last': self.safe_string(ticker, 'last'),
            'previousClose': None,
            'change': None,
            'percentage': None,
            'average': None,
            'baseVolume': self.safe_string(ticker, 'volume'),
            'quoteVolume': None,
            'info': ticker,
        }, market)

    async def fetch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        fetches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :see: https://api.oceanex.pro/doc/v1/#order-book-post
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        market = self.market(symbol)
        request = {
            'market': market['id'],
        }
        if limit is not None:
            request['limit'] = limit
        response = await self.publicGetOrderBook(self.extend(request, params))
        #
        #     {
        #         "code":0,
        #         "message":"Operation successful",
        #         "data": {
        #             "timestamp":1559433057,
        #             "asks": [
        #                 ["100.0","20.0"],
        #                 ["4.74","2000.0"],
        #                 ["1.74","4000.0"],
        #             ],
        #             "bids":[
        #                 ["0.0065","5482873.4"],
        #                 ["0.00649","4781956.2"],
        #                 ["0.00648","2876006.8"],
        #             ],
        #         }
        #     }
        #
        orderbook = self.safe_value(response, 'data', {})
        timestamp = self.safe_timestamp(orderbook, 'timestamp')
        return self.parse_order_book(orderbook, symbol, timestamp)

    async def fetch_order_books(self, symbols: Strings = None, limit: Int = None, params={}):
        """
        fetches information on open orders with bid(buy) and ask(sell) prices, volumes and other data for multiple markets
        :see: https://api.oceanex.pro/doc/v1/#multiple-order-books-post
        :param str[]|None symbols: list of unified market symbols, all symbols fetched if None, default is None
        :param int [limit]: max number of entries per orderbook to return, default is None
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbol
        """
        await self.load_markets()
        if symbols is None:
            symbols = self.symbols
        marketIds = self.market_ids(symbols)
        request = {
            'markets': marketIds,
        }
        if limit is not None:
            request['limit'] = limit
        response = await self.publicGetOrderBookMulti(self.extend(request, params))
        #
        #     {
        #         "code":0,
        #         "message":"Operation successful",
        #         "data": [
        #             {
        #                 "timestamp":1559433057,
        #                 "market": "bagvet",
        #                 "asks": [
        #                     ["100.0","20.0"],
        #                     ["4.74","2000.0"],
        #                     ["1.74","4000.0"],
        #                 ],
        #                 "bids":[
        #                     ["0.0065","5482873.4"],
        #                     ["0.00649","4781956.2"],
        #                     ["0.00648","2876006.8"],
        #                 ],
        #             },
        #             ...,
        #         ],
        #     }
        #
        data = self.safe_value(response, 'data', [])
        result = {}
        for i in range(0, len(data)):
            orderbook = data[i]
            marketId = self.safe_string(orderbook, 'market')
            symbol = self.safe_symbol(marketId)
            timestamp = self.safe_timestamp(orderbook, 'timestamp')
            result[symbol] = self.parse_order_book(orderbook, symbol, timestamp)
        return result

    async def fetch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a particular symbol
        :see: https://api.oceanex.pro/doc/v1/#trades-post
        :param str symbol: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns Trade[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        await self.load_markets()
        market = self.market(symbol)
        request = {
            'market': market['id'],
        }
        if limit is not None:
            request['limit'] = min(limit, 1000)
        response = await self.publicGetTrades(self.extend(request, params))
        #
        #      {
        #          "code":0,
        #          "message":"Operation successful",
        #          "data": [
        #              {
        #                  "id":220247666,
        #                  "price":"3098.62",
        #                  "volume":"0.00196",
        #                  "funds":"6.0732952",
        #                  "market":"ethusdt",
        #                  "created_at":"2022-04-19T19:03:15Z",
        #                  "created_on":1650394995,
        #                  "side":"bid"
        #              },
        #          ]
        #      }
        #
        data = self.safe_list(response, 'data')
        return self.parse_trades(data, market, since, limit)

    def parse_trade(self, trade, market: Market = None) -> Trade:
        #
        # fetchTrades(public)
        #
        #      {
        #          "id":220247666,
        #          "price":"3098.62",
        #          "volume":"0.00196",
        #          "funds":"6.0732952",
        #          "market":"ethusdt",
        #          "created_at":"2022-04-19T19:03:15Z",
        #          "created_on":1650394995,
        #          "side":"bid"
        #      }
        #
        side = self.safe_value(trade, 'side')
        if side == 'bid':
            side = 'buy'
        elif side == 'ask':
            side = 'sell'
        marketId = self.safe_value(trade, 'market')
        symbol = self.safe_symbol(marketId, market)
        timestamp = self.safe_timestamp(trade, 'created_on')
        if timestamp is None:
            timestamp = self.parse8601(self.safe_string(trade, 'created_at'))
        priceString = self.safe_string(trade, 'price')
        amountString = self.safe_string(trade, 'volume')
        return self.safe_trade({
            'info': trade,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': symbol,
            'id': self.safe_string(trade, 'id'),
            'order': None,
            'type': 'limit',
            'takerOrMaker': None,
            'side': side,
            'price': priceString,
            'amount': amountString,
            'cost': None,
            'fee': None,
        }, market)

    async def fetch_time(self, params={}):
        """
        fetches the current integer timestamp in milliseconds from the exchange server
        :see: https://api.oceanex.pro/doc/v1/#api-server-time-post
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int: the current integer timestamp in milliseconds from the exchange server
        """
        response = await self.publicGetTimestamp(params)
        #
        #     {"code":0,"message":"Operation successful","data":1559433420}
        #
        return self.safe_timestamp(response, 'data')

    async def fetch_trading_fees(self, params={}) -> TradingFees:
        """
        fetch the trading fees for multiple markets
        :see: https://api.oceanex.pro/doc/v1/#trading-fees-post
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a dictionary of `fee structures <https://docs.ccxt.com/#/?id=fee-structure>` indexed by market symbols
        """
        response = await self.publicGetFeesTrading(params)
        data = self.safe_value(response, 'data', [])
        result = {}
        for i in range(0, len(data)):
            group = data[i]
            maker = self.safe_value(group, 'ask_fee', {})
            taker = self.safe_value(group, 'bid_fee', {})
            marketId = self.safe_string(group, 'market')
            symbol = self.safe_symbol(marketId)
            result[symbol] = {
                'info': group,
                'symbol': symbol,
                'maker': self.safe_number(maker, 'value'),
                'taker': self.safe_number(taker, 'value'),
                'percentage': True,
            }
        return result

    async def fetch_key(self, params={}):
        response = await self.privateGetKey(params)
        return self.safe_value(response, 'data')

    def parse_balance(self, response) -> Balances:
        data = self.safe_value(response, 'data')
        balances = self.safe_value(data, 'accounts', [])
        result = {'info': response}
        for i in range(0, len(balances)):
            balance = balances[i]
            currencyId = self.safe_value(balance, 'currency')
            code = self.safe_currency_code(currencyId)
            account = self.account()
            account['free'] = self.safe_string(balance, 'balance')
            account['used'] = self.safe_string(balance, 'locked')
            result[code] = account
        return self.safe_balance(result)

    async def fetch_balance(self, params={}) -> Balances:
        """
        query for balance and get the amount of funds available for trading or funds locked in orders
        :see: https://api.oceanex.pro/doc/v1/#account-info-post
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `balance structure <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        await self.load_markets()
        response = await self.privateGetMembersMe(params)
        return self.parse_balance(response)

    async def create_order(self, symbol: str, type: OrderType, side: OrderSide, amount: float, price: Num = None, params={}):
        """
        create a trade order
        :see: https://api.oceanex.pro/doc/v1/#new-order-post
        :param str symbol: unified symbol of the market to create an order in
        :param str type: 'market' or 'limit'
        :param str side: 'buy' or 'sell'
        :param float amount: how much of currency you want to trade in units of base currency
        :param float [price]: the price at which the order is to be fullfilled, in units of the quote currency, ignored in market orders
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        await self.load_markets()
        market = self.market(symbol)
        request = {
            'market': market['id'],
            'side': side,
            'ord_type': type,
            'volume': self.amount_to_precision(symbol, amount),
        }
        if type == 'limit':
            request['price'] = self.price_to_precision(symbol, price)
        response = await self.privatePostOrders(self.extend(request, params))
        data = self.safe_dict(response, 'data')
        return self.parse_order(data, market)

    async def fetch_order(self, id: str, symbol: Str = None, params={}):
        """
        fetches information on an order made by the user
        :see: https://api.oceanex.pro/doc/v1/#order-status-get
        :param str symbol: unified symbol of the market the order was made in
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        await self.load_markets()
        market = None
        if symbol is not None:
            market = self.market(symbol)
        ids = [id]
        request = {'ids': ids}
        response = await self.privateGetOrders(self.extend(request, params))
        data = self.safe_value(response, 'data')
        dataLength = len(data)
        if data is None:
            raise OrderNotFound(self.id + ' could not found matching order')
        if isinstance(id, list):
            orders = self.parse_orders(data, market)
            return orders[0]
        if dataLength == 0:
            raise OrderNotFound(self.id + ' could not found matching order')
        return self.parse_order(data[0], market)

    async def fetch_open_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        fetch all unfilled currently open orders
        :see: https://api.oceanex.pro/doc/v1/#order-status-get
        :param str symbol: unified market symbol
        :param int [since]: the earliest time in ms to fetch open orders for
        :param int [limit]: the maximum number of  open orders structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns Order[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        request = {
            'states': ['wait'],
        }
        return await self.fetch_orders(symbol, since, limit, self.extend(request, params))

    async def fetch_closed_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        fetches information on multiple closed orders made by the user
        :see: https://api.oceanex.pro/doc/v1/#order-status-get
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns Order[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        request = {
            'states': ['done', 'cancel'],
        }
        return await self.fetch_orders(symbol, since, limit, self.extend(request, params))

    async def fetch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        fetches information on multiple orders made by the user
        :see: https://api.oceanex.pro/doc/v1/#order-status-with-filters-post
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns Order[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        if symbol is None:
            raise ArgumentsRequired(self.id + ' fetchOrders() requires a symbol argument')
        await self.load_markets()
        market = self.market(symbol)
        states = self.safe_value(params, 'states', ['wait', 'done', 'cancel'])
        query = self.omit(params, 'states')
        request = {
            'market': market['id'],
            'states': states,
            'need_price': 'True',
        }
        if limit is not None:
            request['limit'] = limit
        response = await self.privateGetOrdersFilter(self.extend(request, query))
        data = self.safe_value(response, 'data', [])
        result = []
        for i in range(0, len(data)):
            orders = self.safe_value(data[i], 'orders', [])
            status = self.parse_order_status(self.safe_value(data[i], 'state'))
            parsedOrders = self.parse_orders(orders, market, since, limit, {'status': status})
            result = self.array_concat(result, parsedOrders)
        return result

    def parse_ohlcv(self, ohlcv, market: Market = None) -> list:
        # [
        #    1559232000,
        #    8889.22,
        #    9028.52,
        #    8889.22,
        #    9028.52
        #    0.3121
        # ]
        return [
            self.safe_timestamp(ohlcv, 0),
            self.safe_number(ohlcv, 1),
            self.safe_number(ohlcv, 2),
            self.safe_number(ohlcv, 3),
            self.safe_number(ohlcv, 4),
            self.safe_number(ohlcv, 5),
        ]

    async def fetch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        fetches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :see: https://api.oceanex.pro/doc/v1/#k-line-post
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        await self.load_markets()
        market = self.market(symbol)
        request = {
            'market': market['id'],
            'period': self.safe_string(self.timeframes, timeframe, timeframe),
        }
        if since is not None:
            request['timestamp'] = since
        if limit is not None:
            request['limit'] = min(limit, 10000)
        response = await self.publicPostK(self.extend(request, params))
        ohlcvs = self.safe_list(response, 'data', [])
        return self.parse_ohlcvs(ohlcvs, market, timeframe, since, limit)

    def parse_order(self, order, market: Market = None) -> Order:
        #
        #     {
        #         "created_at": "2019-01-18T00:38:18Z",
        #         "trades_count": 0,
        #         "remaining_volume": "0.2",
        #         "price": "1001.0",
        #         "created_on": "1547771898",
        #         "side": "buy",
        #         "volume": "0.2",
        #         "state": "wait",
        #         "ord_type": "limit",
        #         "avg_price": "0.0",
        #         "executed_volume": "0.0",
        #         "id": 473797,
        #         "market": "veteth"
        #     }
        #
        status = self.parse_order_status(self.safe_value(order, 'state'))
        marketId = self.safe_string_2(order, 'market', 'market_id')
        symbol = self.safe_symbol(marketId, market)
        timestamp = self.safe_timestamp(order, 'created_on')
        if timestamp is None:
            timestamp = self.parse8601(self.safe_string(order, 'created_at'))
        price = self.safe_string(order, 'price')
        average = self.safe_string(order, 'avg_price')
        amount = self.safe_string(order, 'volume')
        remaining = self.safe_string(order, 'remaining_volume')
        filled = self.safe_string(order, 'executed_volume')
        return self.safe_order({
            'info': order,
            'id': self.safe_string(order, 'id'),
            'clientOrderId': None,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'lastTradeTimestamp': None,
            'symbol': symbol,
            'type': self.safe_value(order, 'ord_type'),
            'timeInForce': None,
            'postOnly': None,
            'side': self.safe_value(order, 'side'),
            'price': price,
            'stopPrice': None,
            'triggerPrice': None,
            'average': average,
            'amount': amount,
            'remaining': remaining,
            'filled': filled,
            'status': status,
            'cost': None,
            'trades': None,
            'fee': None,
        }, market)

    def parse_order_status(self, status):
        statuses = {
            'wait': 'open',
            'done': 'closed',
            'cancel': 'canceled',
        }
        return self.safe_string(statuses, status, status)

    async def cancel_order(self, id: str, symbol: Str = None, params={}):
        """
        cancels an open order
        :see: https://api.oceanex.pro/doc/v1/#cancel-order-post
        :param str id: order id
        :param str symbol: not used by oceanex cancelOrder()
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        await self.load_markets()
        response = await self.privatePostOrderDelete(self.extend({'id': id}, params))
        data = self.safe_dict(response, 'data')
        return self.parse_order(data)

    async def cancel_orders(self, ids, symbol: Str = None, params={}):
        """
        cancel multiple orders
        :see: https://api.oceanex.pro/doc/v1/#cancel-multiple-orders-post
        :param str[] ids: order ids
        :param str symbol: not used by oceanex cancelOrders()
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        await self.load_markets()
        response = await self.privatePostOrderDeleteMulti(self.extend({'ids': ids}, params))
        data = self.safe_list(response, 'data')
        return self.parse_orders(data)

    async def cancel_all_orders(self, symbol: Str = None, params={}):
        """
        cancel all open orders
        :see: https://api.oceanex.pro/doc/v1/#cancel-all-orders-post
        :param str symbol: unified market symbol, only orders in the market of self symbol are cancelled when symbol is not None
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        await self.load_markets()
        response = await self.privatePostOrdersClear(params)
        data = self.safe_list(response, 'data')
        return self.parse_orders(data)

    def sign(self, path, api='public', method='GET', params={}, headers=None, body=None):
        url = self.urls['api']['rest'] + '/' + self.version + '/' + self.implode_params(path, params)
        query = self.omit(params, self.extract_params(path))
        if api == 'public':
            if path == 'tickers_multi' or path == 'order_book/multi':
                request = '?'
                markets = self.safe_value(params, 'markets')
                for i in range(0, len(markets)):
                    request += 'markets[]=' + markets[i] + '&'
                limit = self.safe_value(params, 'limit')
                if limit is not None:
                    request += 'limit=' + limit
                url += request
            elif query:
                url += '?' + self.urlencode(query)
        elif api == 'private':
            self.check_required_credentials()
            request = {
                'uid': self.apiKey,
                'data': query,
            }
            # to set the private key:
            # fs = require('fs')
            # exchange.secret = fs.readFileSync('oceanex.pem', 'utf8')
            jwt_token = self.jwt(request, self.encode(self.secret), 'sha256', True)
            url += '?user_jwt=' + jwt_token
        headers = {'Content-Type': 'application/json'}
        return {'url': url, 'method': method, 'body': body, 'headers': headers}

    def handle_errors(self, code, reason, url, method, headers, body, response, requestHeaders, requestBody):
        #
        #     {"code":1011,"message":"This IP 'x.x.x.x' is not allowed","data":{}}
        #
        if response is None:
            return None
        errorCode = self.safe_string(response, 'code')
        message = self.safe_string(response, 'message')
        if (errorCode is not None) and (errorCode != '0'):
            feedback = self.id + ' ' + body
            self.throw_exactly_matched_exception(self.exceptions['codes'], errorCode, feedback)
            self.throw_exactly_matched_exception(self.exceptions['exact'], message, feedback)
            raise ExchangeError(feedback)
        return None



import os
import sys

root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
sys.path.append(root)

# ----------------------------------------------------------------------------

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

# ----------------------------------------------------------------------------
# -*- coding: utf-8 -*-

from ccxt.base.decimal_to_precision import TICK_SIZE  # noqa E402
import numbers  # noqa E402
from ccxt.base.precise import Precise  # noqa E402
from ccxt.base.errors import OnMaintenance  # noqa E402
from ccxt.base.errors import OperationFailed  # noqa E402

def log_template(exchange, method, entry):
    return ' <<< ' + exchange.id + ' ' + method + ' ::: ' + exchange.json(entry) + ' >>> '


def is_temporary_failure(e):
    return (isinstance(e, OperationFailed)) and (not (isinstance(e, OnMaintenance)))


def string_value(value):
    string_val = None
    if isinstance(value, str):
        string_val = value
    elif value is None:
        string_val = 'undefined'
    else:
        string_val = str(value)
    return string_val


def assert_type(exchange, skipped_properties, entry, key, format):
    if key in skipped_properties:
        return None
    # because "typeof" string is not transpilable without === 'name', we list them manually at this moment
    entry_key_val = exchange.safe_value(entry, key)
    format_key_val = exchange.safe_value(format, key)
    same_string = (isinstance(entry_key_val, str)) and (isinstance(format_key_val, str))
    same_numeric = (isinstance(entry_key_val, numbers.Real)) and (isinstance(format_key_val, numbers.Real))
    same_boolean = ((entry_key_val) or (entry_key_val is False)) and ((format_key_val) or (format_key_val is False))
    same_array = isinstance(entry_key_val, list) and isinstance(format_key_val, list)
    same_object = (isinstance(entry_key_val, dict)) and (isinstance(format_key_val, dict))
    result = (entry_key_val is None) or same_string or same_numeric or same_boolean or same_array or same_object
    return result


def assert_structure(exchange, skipped_properties, method, entry, format, empty_allowed_for=[]):
    log_text = log_template(exchange, method, entry)
    assert entry, 'item is null/undefined' + log_text
    # get all expected & predefined keys for this specific item and ensure thos ekeys exist in parsed structure
    if isinstance(format, list):
        assert isinstance(entry, list), 'entry is not an array' + log_text
        real_length = len(entry)
        expected_length = len(format)
        assert real_length == expected_length, 'entry length is not equal to expected length of ' + str(expected_length) + log_text
        for i in range(0, len(format)):
            empty_allowed_for_this_key = exchange.in_array(i, empty_allowed_for)
            value = entry[i]
            if i in skipped_properties:
                continue
            # check when:
            # - it's not inside "allowe empty values" list
            # - it's not undefined
            if empty_allowed_for_this_key and (value is None):
                continue
            assert value is not None, str(i) + ' index is expected to have a value' + log_text
            # because of other langs, this is needed for arrays
            type_assertion = assert_type(exchange, skipped_properties, entry, i, format)
            assert type_assertion, str(i) + ' index does not have an expected type ' + log_text
    else:
        assert isinstance(entry, dict), 'entry is not an object' + log_text
        keys = list(format.keys())
        for i in range(0, len(keys)):
            key = keys[i]
            if key in skipped_properties:
                continue
            assert key in entry, '\"' + string_value(key) + '\" key is missing from structure' + log_text
            if key in skipped_properties:
                continue
            empty_allowed_for_this_key = exchange.in_array(key, empty_allowed_for)
            value = entry[key]
            # check when:
            # - it's not inside "allowe empty values" list
            # - it's not undefined
            if empty_allowed_for_this_key and (value is None):
                continue
            # if it was in needed keys, then it should have value.
            assert value is not None, '\"' + string_value(key) + '\" key is expected to have a value' + log_text
            # add exclusion for info key, as it can be any type
            if key != 'info':
                type_assertion = assert_type(exchange, skipped_properties, entry, key, format)
                assert type_assertion, '\"' + string_value(key) + '\" key is neither undefined, neither of expected type' + log_text


def assert_timestamp(exchange, skipped_properties, method, entry, now_to_check=None, key_name_or_index='timestamp'):
    log_text = log_template(exchange, method, entry)
    skip_value = exchange.safe_value(skipped_properties, key_name_or_index)
    if skip_value is not None:
        return   # skipped
    is_date_time_object = isinstance(key_name_or_index, str)
    if is_date_time_object:
        assert (key_name_or_index in entry), 'timestamp key \"' + key_name_or_index + '\" is missing from structure' + log_text
    else:
        # if index was provided (mostly from fetchOHLCV) then we check if it exists, as mandatory
        assert not (entry[key_name_or_index] is None), 'timestamp index ' + string_value(key_name_or_index) + ' is undefined' + log_text
    ts = entry[key_name_or_index]
    if ts is not None:
        assert isinstance(ts, numbers.Real), 'timestamp is not numeric' + log_text
        assert isinstance(ts, int), 'timestamp should be an integer' + log_text
        min_ts = 1230940800000  # 03 Jan 2009 - first block
        max_ts = 2147483648000  # 19 Jan 2038 - max int
        assert ts > min_ts, 'timestamp is impossible to be before ' + str(min_ts) + ' (03.01.2009)' + log_text  # 03 Jan 2009 - first block
        assert ts < max_ts, 'timestamp more than ' + str(max_ts) + ' (19.01.2038)' + log_text  # 19 Jan 2038 - int32 overflows # 7258118400000  -> Jan 1 2200
        if now_to_check is not None:
            max_ms_offset = 60000  # 1 min
            assert ts < now_to_check + max_ms_offset, 'returned item timestamp (' + exchange.iso8601(ts) + ') is ahead of the current time (' + exchange.iso8601(now_to_check) + ')' + log_text


def assert_timestamp_and_datetime(exchange, skipped_properties, method, entry, now_to_check=None, key_name_or_index='timestamp'):
    log_text = log_template(exchange, method, entry)
    skip_value = exchange.safe_value(skipped_properties, key_name_or_index)
    if skip_value is not None:
        return
    assert_timestamp(exchange, skipped_properties, method, entry, now_to_check, key_name_or_index)
    is_date_time_object = isinstance(key_name_or_index, str)
    # only in case if the entry is a dictionary, thus it must have 'timestamp' & 'datetime' string keys
    if is_date_time_object:
        # we also test 'datetime' here because it's certain sibling of 'timestamp'
        assert ('datetime' in entry), '\"datetime\" key is missing from structure' + log_text
        dt = entry['datetime']
        if dt is not None:
            assert isinstance(dt, str), '\"datetime\" key does not have a string value' + log_text
            # there are exceptional cases, like getting microsecond-targeted string '2022-08-08T22:03:19.014680Z', so parsed unified timestamp, which carries only 13 digits (millisecond precision) can not be stringified back to microsecond accuracy, causing the bellow assertion to fail
            #    assert (dt === exchange.iso8601 (entry['timestamp']))
            # so, we have to compare with millisecond accururacy
            dt_parsed = exchange.parse8601(dt)
            assert exchange.iso8601(dt_parsed) == exchange.iso8601(entry['timestamp']), 'datetime is not iso8601 of timestamp' + log_text


def assert_currency_code(exchange, skipped_properties, method, entry, actual_code, expected_code=None):
    if ('currency' in skipped_properties) or ('currencyIdAndCode' in skipped_properties):
        return
    log_text = log_template(exchange, method, entry)
    if actual_code is not None:
        assert isinstance(actual_code, str), 'currency code should be either undefined or a string' + log_text
        assert (actual_code in exchange.currencies), 'currency code (\"' + actual_code + '\") should be present in exchange.currencies' + log_text
        if expected_code is not None:
            assert actual_code == expected_code, 'currency code in response (\"' + string_value(actual_code) + '\") should be equal to expected code (\"' + string_value(expected_code) + '\")' + log_text


def assert_valid_currency_id_and_code(exchange, skipped_properties, method, entry, currency_id, currency_code):
    # this is exclusive exceptional key name to be used in `skip-tests.json`, to skip check for currency id and code
    if ('currency' in skipped_properties) or ('currencyIdAndCode' in skipped_properties):
        return
    log_text = log_template(exchange, method, entry)
    undefined_values = currency_id is None and currency_code is None
    defined_values = currency_id is not None and currency_code is not None
    assert undefined_values or defined_values, 'currencyId and currencyCode should be either both defined or both undefined' + log_text
    if defined_values:
        # check by code
        currency_by_code = exchange.currency(currency_code)
        assert currency_by_code['id'] == currency_id, 'currencyId \"' + string_value(currency_id) + '\" does not match currency of code: \"' + string_value(currency_code) + '\"' + log_text
        # check by id
        currency_by_id = exchange.safe_currency(currency_id)
        assert currency_by_id['code'] == currency_code, 'currencyCode ' + string_value(currency_code) + ' does not match currency of id: ' + string_value(currency_id) + log_text


def assert_symbol(exchange, skipped_properties, method, entry, key, expected_symbol=None):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    actual_symbol = exchange.safe_string(entry, key)
    if actual_symbol is not None:
        assert isinstance(actual_symbol, str), 'symbol should be either undefined or a string' + log_text
    if expected_symbol is not None:
        assert actual_symbol == expected_symbol, 'symbol in response (\"' + string_value(actual_symbol) + '\") should be equal to expected symbol (\"' + string_value(expected_symbol) + '\")' + log_text


def assert_symbol_in_markets(exchange, skipped_properties, method, symbol):
    log_text = log_template(exchange, method, {})
    assert (symbol in exchange.markets), 'symbol should be present in exchange.symbols' + log_text


def assert_greater(exchange, skipped_properties, method, entry, key, compare_to):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    value = exchange.safe_string(entry, key)
    if value is not None:
        assert Precise.string_gt(value, compare_to), string_value(key) + ' key (with a value of ' + string_value(value) + ') was expected to be > ' + string_value(compare_to) + log_text


def assert_greater_or_equal(exchange, skipped_properties, method, entry, key, compare_to):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    value = exchange.safe_string(entry, key)
    if value is not None and compare_to is not None:
        assert Precise.string_ge(value, compare_to), string_value(key) + ' key (with a value of ' + string_value(value) + ') was expected to be >= ' + string_value(compare_to) + log_text


def assert_less(exchange, skipped_properties, method, entry, key, compare_to):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    value = exchange.safe_string(entry, key)
    if value is not None and compare_to is not None:
        assert Precise.string_lt(value, compare_to), string_value(key) + ' key (with a value of ' + string_value(value) + ') was expected to be < ' + string_value(compare_to) + log_text


def assert_less_or_equal(exchange, skipped_properties, method, entry, key, compare_to):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    value = exchange.safe_string(entry, key)
    if value is not None and compare_to is not None:
        assert Precise.string_le(value, compare_to), string_value(key) + ' key (with a value of ' + string_value(value) + ') was expected to be <= ' + string_value(compare_to) + log_text


def assert_equal(exchange, skipped_properties, method, entry, key, compare_to):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    value = exchange.safe_string(entry, key)
    if value is not None and compare_to is not None:
        assert Precise.string_eq(value, compare_to), string_value(key) + ' key (with a value of ' + string_value(value) + ') was expected to be equal to ' + string_value(compare_to) + log_text


def assert_non_equal(exchange, skipped_properties, method, entry, key, compare_to):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    value = exchange.safe_string(entry, key)
    if value is not None:
        assert not Precise.string_eq(value, compare_to), string_value(key) + ' key (with a value of ' + string_value(value) + ') was expected not to be equal to ' + string_value(compare_to) + log_text


def assert_in_array(exchange, skipped_properties, method, entry, key, expected_array):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    value = exchange.safe_value(entry, key)
    # todo: remove undefined check
    if value is not None:
        stingified_array_value = exchange.json(expected_array)  # don't use expectedArray.join (','), as it bugs in other languages, if values are bool, undefined or etc..
        assert exchange.in_array(value, expected_array), '\"' + string_value(key) + '\" key (value \"' + string_value(value) + '\") is not from the expected list : [' + stingified_array_value + ']' + log_text


def assert_fee_structure(exchange, skipped_properties, method, entry, key):
    log_text = log_template(exchange, method, entry)
    key_string = string_value(key)
    if isinstance(key, int):
        assert isinstance(entry, list), 'fee container is expected to be an array' + log_text
        assert key < len(entry), 'fee key ' + key_string + ' was expected to be present in entry' + log_text
    else:
        assert isinstance(entry, dict), 'fee container is expected to be an object' + log_text
        assert key in entry, 'fee key \"' + key + '\" was expected to be present in entry' + log_text
    fee_object = exchange.safe_value(entry, key)
    # todo: remove undefined check to make stricter
    if fee_object is not None:
        assert 'cost' in fee_object, key_string + ' fee object should contain \"cost\" key' + log_text
        # assertGreaterOrEqual (exchange, skippedProperties, method, feeObject, 'cost', '0'); # fee might be negative in the case of a rebate or reward
        assert 'currency' in fee_object, '\"' + key_string + '\" fee object should contain \"currency\" key' + log_text
        assert_currency_code(exchange, skipped_properties, method, entry, fee_object['currency'])


def assert_timestamp_order(exchange, method, code_or_symbol, items, ascending=True):
    for i in range(0, len(items)):
        if i > 0:
            current_ts = items[i - 1]['timestamp']
            next_ts = items[i]['timestamp']
            if current_ts is not None and next_ts is not None:
                ascending_or_descending = 'ascending' if ascending else 'descending'
                comparison = (current_ts <= next_ts) if ascending else (current_ts >= next_ts)
                assert comparison, exchange.id + ' ' + method + ' ' + string_value(code_or_symbol) + ' must return a ' + ascending_or_descending + ' sorted array of items by timestamp, but ' + str(current_ts) + ' is opposite with its next ' + str(next_ts) + ' ' + exchange.json(items)


def assert_integer(exchange, skipped_properties, method, entry, key):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    if entry is not None:
        value = exchange.safe_value(entry, key)
        if value is not None:
            is_integer = isinstance(value, int)
            assert is_integer, '\"' + string_value(key) + '\" key (value \"' + string_value(value) + '\") is not an integer' + log_text


def check_precision_accuracy(exchange, skipped_properties, method, entry, key):
    if key in skipped_properties:
        return
    if exchange.is_tick_precision():
        # TICK_SIZE should be above zero
        assert_greater(exchange, skipped_properties, method, entry, key, '0')
        # the below array of integers are inexistent tick-sizes (theoretically technically possible, but not in real-world cases), so their existence in our case indicates to incorrectly implemented tick-sizes, which might mistakenly be implemented with DECIMAL_PLACES, so we throw error
        decimal_numbers = ['2', '3', '4', '6', '7', '8', '9', '11', '12', '13', '14', '15', '16']
        for i in range(0, len(decimal_numbers)):
            num = decimal_numbers[i]
            num_str = num
            assert_non_equal(exchange, skipped_properties, method, entry, key, num_str)
    else:
        assert_integer(exchange, skipped_properties, method, entry, key)  # should be integer
        assert_less_or_equal(exchange, skipped_properties, method, entry, key, '18')  # should be under 18 decimals
        assert_greater_or_equal(exchange, skipped_properties, method, entry, key, '-8')  # in real-world cases, there would not be less than that


def remove_proxy_options(exchange, skipped_properties):
    proxy_url = exchange.check_proxy_url_settings()
    [http_proxy, https_proxy, socks_proxy] = exchange.check_proxy_settings()
    # because of bug in transpiled, about `.proxyUrl` being transpiled into `.proxy_url`, we have to use this workaround
    exchange.set_property(exchange, 'proxyUrl', None)
    exchange.set_property(exchange, 'proxy_url', None)
    exchange.set_property(exchange, 'httpProxy', None)
    exchange.set_property(exchange, 'http_proxy', None)
    exchange.set_property(exchange, 'httpsProxy', None)
    exchange.set_property(exchange, 'https_proxy', None)
    exchange.set_property(exchange, 'socksProxy', None)
    exchange.set_property(exchange, 'socks_proxy', None)
    return [proxy_url, http_proxy, https_proxy, socks_proxy]


def set_proxy_options(exchange, skipped_properties, proxy_url, http_proxy, https_proxy, socks_proxy):
    exchange.proxy_url = proxy_url
    exchange.http_proxy = http_proxy
    exchange.https_proxy = https_proxy
    exchange.socks_proxy = socks_proxy


def assert_non_emtpy_array(exchange, skipped_properties, method, entry, hint=None):
    log_text = log_template(exchange, method, entry)
    if hint is not None:
        log_text = log_text + ' ' + hint
    assert isinstance(entry, list), 'response is expected to be an array' + log_text
    if not ('emptyResponse' in skipped_properties):
        return
    assert len(entry) > 0, 'response is expected to be a non-empty array' + log_text + ' (add \"emptyResponse\" in skip-tests.json to skip this check)'


def assert_round_minute_timestamp(exchange, skipped_properties, method, entry, key):
    if key in skipped_properties:
        return
    log_text = log_template(exchange, method, entry)
    ts = exchange.safe_string(entry, key)
    assert Precise.string_mod(ts, '60000') == '0', 'timestamp should be a multiple of 60 seconds (1 minute)' + log_text



from importlib import (
    import_module,
)
from typing import (
    Any,
)


def import_string(dotted_path: str) -> Any:
    """
    Import a variable using its path and name.

    :param dotted_path: dotted module path and variable/class name
    :return: the attribute/class designated by the last name in the path
    :raise: ImportError, if the import failed

    Source: django.utils.module_loading
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError:
        msg = f"{dotted_path} doesn't look like a module path"
        raise ImportError(msg)

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = f'Module "{module_path}" does not define a "{class_name}" attribute/class'
        raise ImportError(msg)



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_public_currency = publicGetPublicCurrency = Entry('public/currency', 'public', 'GET', {'cost': 10})
    public_get_public_currency_currency = publicGetPublicCurrencyCurrency = Entry('public/currency/{currency}', 'public', 'GET', {'cost': 10})
    public_get_public_symbol = publicGetPublicSymbol = Entry('public/symbol', 'public', 'GET', {'cost': 10})
    public_get_public_symbol_symbol = publicGetPublicSymbolSymbol = Entry('public/symbol/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_ticker = publicGetPublicTicker = Entry('public/ticker', 'public', 'GET', {'cost': 10})
    public_get_public_ticker_symbol = publicGetPublicTickerSymbol = Entry('public/ticker/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_price_rate = publicGetPublicPriceRate = Entry('public/price/rate', 'public', 'GET', {'cost': 10})
    public_get_public_price_history = publicGetPublicPriceHistory = Entry('public/price/history', 'public', 'GET', {'cost': 10})
    public_get_public_price_ticker = publicGetPublicPriceTicker = Entry('public/price/ticker', 'public', 'GET', {'cost': 10})
    public_get_public_price_ticker_symbol = publicGetPublicPriceTickerSymbol = Entry('public/price/ticker/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_trades = publicGetPublicTrades = Entry('public/trades', 'public', 'GET', {'cost': 10})
    public_get_public_trades_symbol = publicGetPublicTradesSymbol = Entry('public/trades/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_orderbook = publicGetPublicOrderbook = Entry('public/orderbook', 'public', 'GET', {'cost': 10})
    public_get_public_orderbook_symbol = publicGetPublicOrderbookSymbol = Entry('public/orderbook/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_candles = publicGetPublicCandles = Entry('public/candles', 'public', 'GET', {'cost': 10})
    public_get_public_candles_symbol = publicGetPublicCandlesSymbol = Entry('public/candles/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_converted_candles = publicGetPublicConvertedCandles = Entry('public/converted/candles', 'public', 'GET', {'cost': 10})
    public_get_public_converted_candles_symbol = publicGetPublicConvertedCandlesSymbol = Entry('public/converted/candles/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_futures_info = publicGetPublicFuturesInfo = Entry('public/futures/info', 'public', 'GET', {'cost': 10})
    public_get_public_futures_info_symbol = publicGetPublicFuturesInfoSymbol = Entry('public/futures/info/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_futures_history_funding = publicGetPublicFuturesHistoryFunding = Entry('public/futures/history/funding', 'public', 'GET', {'cost': 10})
    public_get_public_futures_history_funding_symbol = publicGetPublicFuturesHistoryFundingSymbol = Entry('public/futures/history/funding/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_index_price = publicGetPublicFuturesCandlesIndexPrice = Entry('public/futures/candles/index_price', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_index_price_symbol = publicGetPublicFuturesCandlesIndexPriceSymbol = Entry('public/futures/candles/index_price/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_mark_price = publicGetPublicFuturesCandlesMarkPrice = Entry('public/futures/candles/mark_price', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_mark_price_symbol = publicGetPublicFuturesCandlesMarkPriceSymbol = Entry('public/futures/candles/mark_price/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_premium_index = publicGetPublicFuturesCandlesPremiumIndex = Entry('public/futures/candles/premium_index', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_premium_index_symbol = publicGetPublicFuturesCandlesPremiumIndexSymbol = Entry('public/futures/candles/premium_index/{symbol}', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_open_interest = publicGetPublicFuturesCandlesOpenInterest = Entry('public/futures/candles/open_interest', 'public', 'GET', {'cost': 10})
    public_get_public_futures_candles_open_interest_symbol = publicGetPublicFuturesCandlesOpenInterestSymbol = Entry('public/futures/candles/open_interest/{symbol}', 'public', 'GET', {'cost': 10})
    private_get_spot_balance = privateGetSpotBalance = Entry('spot/balance', 'private', 'GET', {'cost': 15})
    private_get_spot_balance_currency = privateGetSpotBalanceCurrency = Entry('spot/balance/{currency}', 'private', 'GET', {'cost': 15})
    private_get_spot_order = privateGetSpotOrder = Entry('spot/order', 'private', 'GET', {'cost': 1})
    private_get_spot_order_client_order_id = privateGetSpotOrderClientOrderId = Entry('spot/order/{client_order_id}', 'private', 'GET', {'cost': 1})
    private_get_spot_fee = privateGetSpotFee = Entry('spot/fee', 'private', 'GET', {'cost': 15})
    private_get_spot_fee_symbol = privateGetSpotFeeSymbol = Entry('spot/fee/{symbol}', 'private', 'GET', {'cost': 15})
    private_get_spot_history_order = privateGetSpotHistoryOrder = Entry('spot/history/order', 'private', 'GET', {'cost': 15})
    private_get_spot_history_trade = privateGetSpotHistoryTrade = Entry('spot/history/trade', 'private', 'GET', {'cost': 15})
    private_get_margin_account = privateGetMarginAccount = Entry('margin/account', 'private', 'GET', {'cost': 1})
    private_get_margin_account_isolated_symbol = privateGetMarginAccountIsolatedSymbol = Entry('margin/account/isolated/{symbol}', 'private', 'GET', {'cost': 1})
    private_get_margin_account_cross_currency = privateGetMarginAccountCrossCurrency = Entry('margin/account/cross/{currency}', 'private', 'GET', {'cost': 1})
    private_get_margin_order = privateGetMarginOrder = Entry('margin/order', 'private', 'GET', {'cost': 1})
    private_get_margin_order_client_order_id = privateGetMarginOrderClientOrderId = Entry('margin/order/{client_order_id}', 'private', 'GET', {'cost': 1})
    private_get_margin_config = privateGetMarginConfig = Entry('margin/config', 'private', 'GET', {'cost': 15})
    private_get_margin_history_order = privateGetMarginHistoryOrder = Entry('margin/history/order', 'private', 'GET', {'cost': 15})
    private_get_margin_history_trade = privateGetMarginHistoryTrade = Entry('margin/history/trade', 'private', 'GET', {'cost': 15})
    private_get_margin_history_positions = privateGetMarginHistoryPositions = Entry('margin/history/positions', 'private', 'GET', {'cost': 15})
    private_get_margin_history_clearing = privateGetMarginHistoryClearing = Entry('margin/history/clearing', 'private', 'GET', {'cost': 15})
    private_get_futures_balance = privateGetFuturesBalance = Entry('futures/balance', 'private', 'GET', {'cost': 15})
    private_get_futures_balance_currency = privateGetFuturesBalanceCurrency = Entry('futures/balance/{currency}', 'private', 'GET', {'cost': 15})
    private_get_futures_account = privateGetFuturesAccount = Entry('futures/account', 'private', 'GET', {'cost': 1})
    private_get_futures_account_isolated_symbol = privateGetFuturesAccountIsolatedSymbol = Entry('futures/account/isolated/{symbol}', 'private', 'GET', {'cost': 1})
    private_get_futures_order = privateGetFuturesOrder = Entry('futures/order', 'private', 'GET', {'cost': 1})
    private_get_futures_order_client_order_id = privateGetFuturesOrderClientOrderId = Entry('futures/order/{client_order_id}', 'private', 'GET', {'cost': 1})
    private_get_futures_config = privateGetFuturesConfig = Entry('futures/config', 'private', 'GET', {'cost': 15})
    private_get_futures_fee = privateGetFuturesFee = Entry('futures/fee', 'private', 'GET', {'cost': 15})
    private_get_futures_fee_symbol = privateGetFuturesFeeSymbol = Entry('futures/fee/{symbol}', 'private', 'GET', {'cost': 15})
    private_get_futures_history_order = privateGetFuturesHistoryOrder = Entry('futures/history/order', 'private', 'GET', {'cost': 15})
    private_get_futures_history_trade = privateGetFuturesHistoryTrade = Entry('futures/history/trade', 'private', 'GET', {'cost': 15})
    private_get_futures_history_positions = privateGetFuturesHistoryPositions = Entry('futures/history/positions', 'private', 'GET', {'cost': 15})
    private_get_futures_history_clearing = privateGetFuturesHistoryClearing = Entry('futures/history/clearing', 'private', 'GET', {'cost': 15})
    private_get_wallet_balance = privateGetWalletBalance = Entry('wallet/balance', 'private', 'GET', {'cost': 30})
    private_get_wallet_balance_currency = privateGetWalletBalanceCurrency = Entry('wallet/balance/{currency}', 'private', 'GET', {'cost': 30})
    private_get_wallet_crypto_address = privateGetWalletCryptoAddress = Entry('wallet/crypto/address', 'private', 'GET', {'cost': 30})
    private_get_wallet_crypto_address_recent_deposit = privateGetWalletCryptoAddressRecentDeposit = Entry('wallet/crypto/address/recent-deposit', 'private', 'GET', {'cost': 30})
    private_get_wallet_crypto_address_recent_withdraw = privateGetWalletCryptoAddressRecentWithdraw = Entry('wallet/crypto/address/recent-withdraw', 'private', 'GET', {'cost': 30})
    private_get_wallet_crypto_address_check_mine = privateGetWalletCryptoAddressCheckMine = Entry('wallet/crypto/address/check-mine', 'private', 'GET', {'cost': 30})
    private_get_wallet_transactions = privateGetWalletTransactions = Entry('wallet/transactions', 'private', 'GET', {'cost': 30})
    private_get_wallet_transactions_tx_id = privateGetWalletTransactionsTxId = Entry('wallet/transactions/{tx_id}', 'private', 'GET', {'cost': 30})
    private_get_wallet_crypto_fee_estimate = privateGetWalletCryptoFeeEstimate = Entry('wallet/crypto/fee/estimate', 'private', 'GET', {'cost': 30})
    private_get_wallet_airdrops = privateGetWalletAirdrops = Entry('wallet/airdrops', 'private', 'GET', {'cost': 30})
    private_get_wallet_amount_locks = privateGetWalletAmountLocks = Entry('wallet/amount-locks', 'private', 'GET', {'cost': 30})
    private_get_sub_account = privateGetSubAccount = Entry('sub-account', 'private', 'GET', {'cost': 15})
    private_get_sub_account_acl = privateGetSubAccountAcl = Entry('sub-account/acl', 'private', 'GET', {'cost': 15})
    private_get_sub_account_balance_subaccid = privateGetSubAccountBalanceSubAccID = Entry('sub-account/balance/{subAccID}', 'private', 'GET', {'cost': 15})
    private_get_sub_account_crypto_address_subaccid_currency = privateGetSubAccountCryptoAddressSubAccIDCurrency = Entry('sub-account/crypto/address/{subAccID}/{currency}', 'private', 'GET', {'cost': 15})
    private_post_spot_order = privatePostSpotOrder = Entry('spot/order', 'private', 'POST', {'cost': 1})
    private_post_spot_order_list = privatePostSpotOrderList = Entry('spot/order/list', 'private', 'POST', {'cost': 1})
    private_post_margin_order = privatePostMarginOrder = Entry('margin/order', 'private', 'POST', {'cost': 1})
    private_post_margin_order_list = privatePostMarginOrderList = Entry('margin/order/list', 'private', 'POST', {'cost': 1})
    private_post_futures_order = privatePostFuturesOrder = Entry('futures/order', 'private', 'POST', {'cost': 1})
    private_post_futures_order_list = privatePostFuturesOrderList = Entry('futures/order/list', 'private', 'POST', {'cost': 1})
    private_post_wallet_crypto_address = privatePostWalletCryptoAddress = Entry('wallet/crypto/address', 'private', 'POST', {'cost': 30})
    private_post_wallet_crypto_withdraw = privatePostWalletCryptoWithdraw = Entry('wallet/crypto/withdraw', 'private', 'POST', {'cost': 30})
    private_post_wallet_convert = privatePostWalletConvert = Entry('wallet/convert', 'private', 'POST', {'cost': 30})
    private_post_wallet_transfer = privatePostWalletTransfer = Entry('wallet/transfer', 'private', 'POST', {'cost': 30})
    private_post_wallet_internal_withdraw = privatePostWalletInternalWithdraw = Entry('wallet/internal/withdraw', 'private', 'POST', {'cost': 30})
    private_post_wallet_crypto_check_offchain_available = privatePostWalletCryptoCheckOffchainAvailable = Entry('wallet/crypto/check-offchain-available', 'private', 'POST', {'cost': 30})
    private_post_wallet_crypto_fees_estimate = privatePostWalletCryptoFeesEstimate = Entry('wallet/crypto/fees/estimate', 'private', 'POST', {'cost': 30})
    private_post_wallet_airdrops_id_claim = privatePostWalletAirdropsIdClaim = Entry('wallet/airdrops/{id}/claim', 'private', 'POST', {'cost': 30})
    private_post_sub_account_freeze = privatePostSubAccountFreeze = Entry('sub-account/freeze', 'private', 'POST', {'cost': 15})
    private_post_sub_account_activate = privatePostSubAccountActivate = Entry('sub-account/activate', 'private', 'POST', {'cost': 15})
    private_post_sub_account_transfer = privatePostSubAccountTransfer = Entry('sub-account/transfer', 'private', 'POST', {'cost': 15})
    private_post_sub_account_acl = privatePostSubAccountAcl = Entry('sub-account/acl', 'private', 'POST', {'cost': 15})
    private_patch_spot_order_client_order_id = privatePatchSpotOrderClientOrderId = Entry('spot/order/{client_order_id}', 'private', 'PATCH', {'cost': 1})
    private_patch_margin_order_client_order_id = privatePatchMarginOrderClientOrderId = Entry('margin/order/{client_order_id}', 'private', 'PATCH', {'cost': 1})
    private_patch_futures_order_client_order_id = privatePatchFuturesOrderClientOrderId = Entry('futures/order/{client_order_id}', 'private', 'PATCH', {'cost': 1})
    private_delete_spot_order = privateDeleteSpotOrder = Entry('spot/order', 'private', 'DELETE', {'cost': 1})
    private_delete_spot_order_client_order_id = privateDeleteSpotOrderClientOrderId = Entry('spot/order/{client_order_id}', 'private', 'DELETE', {'cost': 1})
    private_delete_margin_position = privateDeleteMarginPosition = Entry('margin/position', 'private', 'DELETE', {'cost': 1})
    private_delete_margin_position_isolated_symbol = privateDeleteMarginPositionIsolatedSymbol = Entry('margin/position/isolated/{symbol}', 'private', 'DELETE', {'cost': 1})
    private_delete_margin_order = privateDeleteMarginOrder = Entry('margin/order', 'private', 'DELETE', {'cost': 1})
    private_delete_margin_order_client_order_id = privateDeleteMarginOrderClientOrderId = Entry('margin/order/{client_order_id}', 'private', 'DELETE', {'cost': 1})
    private_delete_futures_position = privateDeleteFuturesPosition = Entry('futures/position', 'private', 'DELETE', {'cost': 1})
    private_delete_futures_position_margin_mode_symbol = privateDeleteFuturesPositionMarginModeSymbol = Entry('futures/position/{margin_mode}/{symbol}', 'private', 'DELETE', {'cost': 1})
    private_delete_futures_order = privateDeleteFuturesOrder = Entry('futures/order', 'private', 'DELETE', {'cost': 1})
    private_delete_futures_order_client_order_id = privateDeleteFuturesOrderClientOrderId = Entry('futures/order/{client_order_id}', 'private', 'DELETE', {'cost': 1})
    private_delete_wallet_crypto_withdraw_id = privateDeleteWalletCryptoWithdrawId = Entry('wallet/crypto/withdraw/{id}', 'private', 'DELETE', {'cost': 30})
    private_put_margin_account_isolated_symbol = privatePutMarginAccountIsolatedSymbol = Entry('margin/account/isolated/{symbol}', 'private', 'PUT', {'cost': 1})
    private_put_futures_account_isolated_symbol = privatePutFuturesAccountIsolatedSymbol = Entry('futures/account/isolated/{symbol}', 'private', 'PUT', {'cost': 1})
    private_put_wallet_crypto_withdraw_id = privatePutWalletCryptoWithdrawId = Entry('wallet/crypto/withdraw/{id}', 'private', 'PUT', {'cost': 30})



from ccxt.base.types import Entry


class ImplicitAPI:
    webexchange_get = webExchangeGet = Entry('', 'webExchange', 'GET', {})
    web_get_rest_api = webGetRestApi = Entry('rest-api', 'web', 'GET', {})
    public_get_v1_symbols = publicGetV1Symbols = Entry('v1/symbols', 'public', 'GET', {'cost': 5})
    public_get_v1_symbols_details_symbol = publicGetV1SymbolsDetailsSymbol = Entry('v1/symbols/details/{symbol}', 'public', 'GET', {'cost': 5})
    public_get_v1_staking_rates = publicGetV1StakingRates = Entry('v1/staking/rates', 'public', 'GET', {'cost': 5})
    public_get_v1_pubticker_symbol = publicGetV1PubtickerSymbol = Entry('v1/pubticker/{symbol}', 'public', 'GET', {'cost': 5})
    public_get_v2_ticker_symbol = publicGetV2TickerSymbol = Entry('v2/ticker/{symbol}', 'public', 'GET', {'cost': 5})
    public_get_v2_candles_symbol_timeframe = publicGetV2CandlesSymbolTimeframe = Entry('v2/candles/{symbol}/{timeframe}', 'public', 'GET', {'cost': 5})
    public_get_v1_trades_symbol = publicGetV1TradesSymbol = Entry('v1/trades/{symbol}', 'public', 'GET', {'cost': 5})
    public_get_v1_auction_symbol = publicGetV1AuctionSymbol = Entry('v1/auction/{symbol}', 'public', 'GET', {'cost': 5})
    public_get_v1_auction_symbol_history = publicGetV1AuctionSymbolHistory = Entry('v1/auction/{symbol}/history', 'public', 'GET', {'cost': 5})
    public_get_v1_pricefeed = publicGetV1Pricefeed = Entry('v1/pricefeed', 'public', 'GET', {'cost': 5})
    public_get_v1_book_symbol = publicGetV1BookSymbol = Entry('v1/book/{symbol}', 'public', 'GET', {'cost': 5})
    public_get_v1_earn_rates = publicGetV1EarnRates = Entry('v1/earn/rates', 'public', 'GET', {'cost': 5})
    private_post_v1_staking_unstake = privatePostV1StakingUnstake = Entry('v1/staking/unstake', 'private', 'POST', {'cost': 1})
    private_post_v1_staking_stake = privatePostV1StakingStake = Entry('v1/staking/stake', 'private', 'POST', {'cost': 1})
    private_post_v1_staking_rewards = privatePostV1StakingRewards = Entry('v1/staking/rewards', 'private', 'POST', {'cost': 1})
    private_post_v1_staking_history = privatePostV1StakingHistory = Entry('v1/staking/history', 'private', 'POST', {'cost': 1})
    private_post_v1_order_new = privatePostV1OrderNew = Entry('v1/order/new', 'private', 'POST', {'cost': 1})
    private_post_v1_order_cancel = privatePostV1OrderCancel = Entry('v1/order/cancel', 'private', 'POST', {'cost': 1})
    private_post_v1_wrap_symbol = privatePostV1WrapSymbol = Entry('v1/wrap/{symbol}', 'private', 'POST', {'cost': 1})
    private_post_v1_order_cancel_session = privatePostV1OrderCancelSession = Entry('v1/order/cancel/session', 'private', 'POST', {'cost': 1})
    private_post_v1_order_cancel_all = privatePostV1OrderCancelAll = Entry('v1/order/cancel/all', 'private', 'POST', {'cost': 1})
    private_post_v1_order_status = privatePostV1OrderStatus = Entry('v1/order/status', 'private', 'POST', {'cost': 1})
    private_post_v1_orders = privatePostV1Orders = Entry('v1/orders', 'private', 'POST', {'cost': 1})
    private_post_v1_mytrades = privatePostV1Mytrades = Entry('v1/mytrades', 'private', 'POST', {'cost': 1})
    private_post_v1_notionalvolume = privatePostV1Notionalvolume = Entry('v1/notionalvolume', 'private', 'POST', {'cost': 1})
    private_post_v1_tradevolume = privatePostV1Tradevolume = Entry('v1/tradevolume', 'private', 'POST', {'cost': 1})
    private_post_v1_clearing_new = privatePostV1ClearingNew = Entry('v1/clearing/new', 'private', 'POST', {'cost': 1})
    private_post_v1_clearing_status = privatePostV1ClearingStatus = Entry('v1/clearing/status', 'private', 'POST', {'cost': 1})
    private_post_v1_clearing_cancel = privatePostV1ClearingCancel = Entry('v1/clearing/cancel', 'private', 'POST', {'cost': 1})
    private_post_v1_clearing_confirm = privatePostV1ClearingConfirm = Entry('v1/clearing/confirm', 'private', 'POST', {'cost': 1})
    private_post_v1_balances = privatePostV1Balances = Entry('v1/balances', 'private', 'POST', {'cost': 1})
    private_post_v1_balances_staking = privatePostV1BalancesStaking = Entry('v1/balances/staking', 'private', 'POST', {'cost': 1})
    private_post_v1_notionalbalances_currency = privatePostV1NotionalbalancesCurrency = Entry('v1/notionalbalances/{currency}', 'private', 'POST', {'cost': 1})
    private_post_v1_transfers = privatePostV1Transfers = Entry('v1/transfers', 'private', 'POST', {'cost': 1})
    private_post_v1_addresses_network = privatePostV1AddressesNetwork = Entry('v1/addresses/{network}', 'private', 'POST', {'cost': 1})
    private_post_v1_deposit_network_newaddress = privatePostV1DepositNetworkNewAddress = Entry('v1/deposit/{network}/newAddress', 'private', 'POST', {'cost': 1})
    private_post_v1_deposit_currency_newaddress = privatePostV1DepositCurrencyNewAddress = Entry('v1/deposit/{currency}/newAddress', 'private', 'POST', {'cost': 1})
    private_post_v1_withdraw_currency = privatePostV1WithdrawCurrency = Entry('v1/withdraw/{currency}', 'private', 'POST', {'cost': 1})
    private_post_v1_account_transfer_currency = privatePostV1AccountTransferCurrency = Entry('v1/account/transfer/{currency}', 'private', 'POST', {'cost': 1})
    private_post_v1_payments_addbank = privatePostV1PaymentsAddbank = Entry('v1/payments/addbank', 'private', 'POST', {'cost': 1})
    private_post_v1_payments_methods = privatePostV1PaymentsMethods = Entry('v1/payments/methods', 'private', 'POST', {'cost': 1})
    private_post_v1_payments_sen_withdraw = privatePostV1PaymentsSenWithdraw = Entry('v1/payments/sen/withdraw', 'private', 'POST', {'cost': 1})
    private_post_v1_balances_earn = privatePostV1BalancesEarn = Entry('v1/balances/earn', 'private', 'POST', {'cost': 1})
    private_post_v1_earn_interest = privatePostV1EarnInterest = Entry('v1/earn/interest', 'private', 'POST', {'cost': 1})
    private_post_v1_earn_history = privatePostV1EarnHistory = Entry('v1/earn/history', 'private', 'POST', {'cost': 1})
    private_post_v1_approvedaddresses_network_request = privatePostV1ApprovedAddressesNetworkRequest = Entry('v1/approvedAddresses/{network}/request', 'private', 'POST', {'cost': 1})
    private_post_v1_approvedaddresses_account_network = privatePostV1ApprovedAddressesAccountNetwork = Entry('v1/approvedAddresses/account/{network}', 'private', 'POST', {'cost': 1})
    private_post_v1_approvedaddresses_network_remove = privatePostV1ApprovedAddressesNetworkRemove = Entry('v1/approvedAddresses/{network}/remove', 'private', 'POST', {'cost': 1})
    private_post_v1_account = privatePostV1Account = Entry('v1/account', 'private', 'POST', {'cost': 1})
    private_post_v1_account_create = privatePostV1AccountCreate = Entry('v1/account/create', 'private', 'POST', {'cost': 1})
    private_post_v1_account_list = privatePostV1AccountList = Entry('v1/account/list', 'private', 'POST', {'cost': 1})
    private_post_v1_heartbeat = privatePostV1Heartbeat = Entry('v1/heartbeat', 'private', 'POST', {'cost': 1})
